==> cui-spa-app-class-0.1.13.tgz: package/modules/router.js <==
export default class ROUTER {

    constructor(oRoutes, sScriptRoot, cAPP) {

        // Full route set
        this.oRoutes = oRoutes;

        // Create a reference to the application class
        this.cAPP = cAPP;

        // Freeze this object so it can naver be changed!
        Object.freeze(this.oRoutes);

        this.sScriptRoot = sScriptRoot;
        this.sScreenRoot = `${sScriptRoot}/screens`;

        this.sOrigin = window.location.origin;

        // Array if obeject contain currently loaded route modules
        this.aoCurrentRoutesModules = null;

        // Placeholder object used when the user requests a new path
        this.oNextRoute = null;

        this.mModuleRegistry = new Map();

        window.onpopstate = (evt) => {
            console.log("On Pop State called!", evt);

            let navState;

            if(evt && evt.state){
                navState = evt.state;
            }

            // Browser popstate already updated location URL to load. Need to call navTo with current location's URL. 
            let sPopHref = document.location.href;
            sPopHref = sPopHref.replace(`${document.location.protocol}//${document.location.hostname}:${document.location.port}`, "");

            this.navTo(sPopHref, null, null, null, navState);
        }
    }

    async loadModule(sModulePath) {

        if (this.mModuleRegistry.has(sModulePath)) {

            return this.mModuleRegistry.get(sModulePath);
        }
        else {

            let cModule = await import(sModulePath);

            this.mModuleRegistry.set(sModulePath, cModule);

            return cModule;
        }

    }

    async updateRoute(oNavState) {
        let bPreventNav = false;


        if (this.oNextRoute) {

            let aoUnloadRoutes = null;
            let asLoadRoutes = null;
            let oPopState = null;

            console.log("updateRoute");
            console.log("Current loaded routes array")
            console.log(this.aoCurrentRoutesModules);

            console.log("Next Routes routes");
            console.log(this.oNextRoute.asRoutePaths);

            // Check to see if we already have modules loaded
            if (this.aoCurrentRoutesModules && this.aoCurrentRoutesModules.length) {

                // Loop throught the routes compare them to what is already loaded
                for (let r = 0, rLen = this.oNextRoute.asRoutePaths.length; r < rLen; r++) {

                    // Validate the current cursor has this index
                    if (this.aoCurrentRoutesModules[r]) {

                        // Check to see if we dont have a match, because that is when we care!
                        if (this.aoCurrentRoutesModules[r].sModulePath !== this.oNextRoute.asRoutePaths[r]) {
                            


                        }
                    }
                    else {

                        // Everything from this point on is new!
                        asLoadRoutes = this.oNextRoute.asRoutePaths.slice(r);

                    }

                    if ((r + 1 === rLen) && (this.aoCurrentRoutesModules.length > rLen)) {

                        console.log("At the end of the requested route, but there are more routes still in memory");

                        aoUnloadRoutes = this.aoCurrentRoutesModules.slice(rLen).reverse();

                        this.aoCurrentRoutesModules = this.aoCurrentRoutesModules.slice(0, rLen);
                    }

                }

            }
            else {

                // Nothing to worry about unloading, so load all new
                asLoadRoutes = this.oNextRoute.asRoutePaths;
            }

            /*
             * Screen to unloaded
             */

            // Check to see if we have a page that is being unloaded
            if (this.cCurrentScreenModule) {

                if (this.cCurrentScreenModule._unload) {

                    bPreventNav = this.cCurrentScreenModule._navGuard();

                    if(!bPreventNav){
                        oPopState = this.cCurrentScreenModule._unload();    
                    }
                }
                else {
                    
                    console.log("There is a screen, but no unload method to call");
                }

            }
            else {

                console.log("No screen module to unload");
            }

            if(!bPreventNav){

                /*
                 * Old route to unload (already in reversed order)
                 */

                // Check for modules that need to be unloaded
                if (aoUnloadRoutes && aoUnloadRoutes.length) {

                    for (let u = 0, uLen = aoUnloadRoutes.length; u < uLen; u++) {
                     
                        console.log(aoUnloadRoutes[u]);

                        if (aoUnloadRoutes[u].cModule._unload) {

                            await aoUnloadRoutes[u].cModule._unload();
                        }
                        
                    }
                }

                /*
                 * New route modules to load
                 */

                // Check for load modules
                if (asLoadRoutes && asLoadRoutes.length) {

                    for (let sRouteModule of asLoadRoutes) {

                        let cModule = await this.loadModule(sRouteModule);

                        let oRouteProps = {};
                        if(this.aoCurrentRoutesModules){
                            oRouteProps.oParentRoute = this.aoCurrentRoutesModules[this.aoCurrentRoutesModules.length - 1].cModule;
                        }
                        oRouteProps.oNavState = oNavState

                        // Build the module object
                        let oModuleLoad = {
                            sModulePath: sRouteModule,
                            cModule: new cModule.default(oRouteProps)
                        };

                        // Initialize the router modules
                        await oModuleLoad.cModule._init();

                        // Initialize the array if it doesnt exist
                        if (!this.aoCurrentRoutesModules) {
                            this.aoCurrentRoutesModules = [];
                        }

                        // Save off the module
                        this.aoCurrentRoutesModules.push(oModuleLoad);
                    }

                }

                /*DEBUG: START*/
                console.log('currentRoutes', this.aoCurrentRoutesModules);
                /*DEBUG: END*/

                /*
                 * Screen to load
                 */
                if (this.oNextRoute.sPagePath) {

                    let cScreen = await this.loadModule(this.oNextRoute.sPagePath);

                    /*DEBUG: START*/
                    console.log('cScreen', cScreen);
                    /*DEBUG: END*/

                    let oScreenProps = {};
                    if(this.aoCurrentRoutesModules){
                        oScreenProps.oParentRoute = this.aoCurrentRoutesModules[this.aoCurrentRoutesModules.length - 1].cModule;
                    }
                   
                    oScreenProps.oNavState = oNavState;
                 

                    this.cCurrentScreenModule = new cScreen.default(oScreenProps, this.cAPP);

                    await this.cCurrentScreenModule._init();

                }
                else {

                    throw new Error(`No new next route screen was defined`);
                }

                // Set history state of current URL with value returned to oPopState. 
                window.history.replaceState(oPopState, "",);

                // Create new history entry with new pages URL. 
                window.history.pushState({}, "", this.oNextRoute.sURLBar);
        
                document.title = this.cCurrentScreenModule.sTitle;
        
                // Clear out the next router info
                this.oNextRoute = null;
            }

        }
        else {

            throw new Error(`No new next route is currently defined`);
        }

    }

    validateRoute(sRequestedRoute, sRequestSource = "manual") {

        // Check to make sure the path is not
        if (!sRequestSource.startsWith('..')) {

            let bWasFlushed = false;

            let sWorkingPath = sRequestedRoute;

            let oNewCursor = this.oRoutes;
            let asRoutePaths = [];
            let sPagePath = null;
            let sURLBar = "/";
            let sFromRoot = "";

            if (sWorkingPath.startsWith('/')) {
                sWorkingPath = sWorkingPath.slice(1);
            }

            while (true) {

                let sBasePath = null;
                let routeModulePath = null;

                if (sFromRoot && sFromRoot.length) {
                    sBasePath = `${this.sScreenRoot}/${sFromRoot}`;
                }
                else {
                    sBasePath = this.sScreenRoot;
                }

                // A route folder request
                if (sWorkingPath.length === 0 || sWorkingPath === "index") {

                    // Check to see if index exists for this route
                    if (oNewCursor.index) {

                        // Check for and add a router module if one exists in this folder
                        if (oNewCursor['@routeModule']) {

                            routeModulePath = `${sBasePath}/${oNewCursor['@routeModule']}`;

                            if (asRoutePaths.indexOf(routeModulePath) === -1) {
                                asRoutePaths.push(routeModulePath);
                            }
                        }

                        // Build the page path
                        sPagePath = `${sBasePath}/${oNewCursor.index}`;

                        sURLBar += "index";

                        break;
                    }
                    else {

                        throw new Error(`Invalid route requested: ${sRequestedRoute}. Requested folder doesnt have a index.`);
                    }

                }
                else {

                    // Check to see if we have a path seperator
                    let iSlashIndex = sWorkingPath.indexOf('/');
                    let sPathPiece = null;

                    if (iSlashIndex === -1) {

                        sPathPiece = sWorkingPath.slice(0);
                        sWorkingPath = "";
                    }
                    else {

                        sPathPiece = sWorkingPath.slice(0, iSlashIndex);
                        sWorkingPath = sWorkingPath.slice(iSlashIndex + 1);
                    }

                    // Check the current path peice
                    if (oNewCursor[sPathPiece]) {

                        // Check to see if this is a page
                        if (typeof oNewCursor[sPathPiece] === "string") {

                            if (sWorkingPath.length === 0) {

                                // Check for and add a router module if one exists in this folder and is not part of the current path
                                if (oNewCursor['@routeModule']) {

                                    routeModulePath = `${sBasePath}/${oNewCursor['@routeModule']}`;

                                    if (asRoutePaths.indexOf(routeModulePath) === -1) {
                                        asRoutePaths.push(routeModulePath);
                                    }
                                }

                                // Build the page path
                                sPagePath = `${sBasePath}/${oNewCursor[sPathPiece]}`;

                                sURLBar += sPathPiece;

                                break;
                            }
                            else {

                                throw new Error(`Invalid route requested: ${sRequestedRoute}. Reached end of cursor, but not path`);
                            }

                        }
                        else {

                            // Check for and add a router module if one exists in this folder and is not part of the current path
                            if (oNewCursor['@routeModule']) {

                                routeModulePath = `${sBasePath}/${oNewCursor['@routeModule']}`;

                                if (asRoutePaths.indexOf(routeModulePath) === -1) {
                                    asRoutePaths.push(routeModulePath);
                                }
                            }

                            // Update from path to adopt the next level
                            if (sFromRoot.length) { 
                                sFromRoot = `${sFromRoot}/${sPathPiece}`;
                            }
                            else {
                                sFromRoot = sPathPiece;
                            }

                            sURLBar += `${sPathPiece}/`;

                            // Advance the cursor
                            oNewCursor = oNewCursor[sPathPiece];

                        }

                    }
                    else {

                        throw new Error(`Invalid route requested: ${sRequestedRoute}. Path cursor does not have a option for: ${sPathPiece}`);
                    }

                }

            }

            this.oNextRoute = {
                asRoutePaths: asRoutePaths,
                sPagePath: sPagePath,
                sURLBar: sURLBar
            }

        }
        else {

            throw new Error(`Route paths must be direct siblings or absolutes from route`);
        }

    }

    async navTo(sDestination, oNavData, eEvent, sSource, oHistoryState) {
        
        console.log(`User requested to navigate to: ${sDestination}`, eEvent, sSource, oNavData);

        let bContinue = true;

        // Start by trying to request the 
        try {

            this.validateRoute(sDestination);
        }
        catch(oError) {

            bContinue = false;
            console.log(oError);
        }

        if (bContinue) {
            
            let oNavState = {};
            if(sDestination){
                oNavState.sDestination = sDestination;
            }
            if(sSource){
                oNavState.sSource = sSource;
            }
            if(oNavData){
                oNavState.oNavData = oNavData;
            }        
            if(oHistoryState){
                oNavState.oHistoryState = oHistoryState;
            }
            

            this.updateRoute(oNavState);
        }

    }
}

==> cui-spa-app-class-0.1.17.tgz: package/modules/router.js <==
export default class ROUTER {

    constructor(oRoutes, sScriptRoot, cAPP) {

        // Full route set
        this.oRoutes = oRoutes;

        // Create a reference to the application class
        this.cAPP = cAPP;

        // Freeze this object so it can naver be changed!
        Object.freeze(this.oRoutes);

        this.sScriptRoot = sScriptRoot;
        this.sScreenRoot = `${sScriptRoot}/screens`;

        this.sOrigin = window.location.origin;

        // Array if obeject contain currently loaded route modules
        this.aoCurrentRoutesModules = null;

        // Placeholder object used when the user requests a new path
        this.oNextRoute = null;

        this.mModuleRegistry = new Map();

        window.onpopstate = (evt) => {
            console.log("On Pop State called!", evt);

            let navState;

            if(evt && evt.state){
                navState = evt.state;
            }

            // Browser popstate already updated location URL to load. Need to call navTo with current location's URL. 
            let sPopHref = document.location.href;
            sPopHref = sPopHref.replace(`${document.location.protocol}//${document.location.hostname}:${document.location.port}`, "");

            this.navTo(sPopHref, null, null, null, navState);
        }
    }

    async loadModule(sModulePath) {

        if (this.mModuleRegistry.has(sModulePath)) {

            return this.mModuleRegistry.get(sModulePath);
        }
        else {

            let cModule = await import(sModulePath);

            this.mModuleRegistry.set(sModulePath, cModule);

            return cModule;
        }

    }

    async updateRoute(oNavState) {
        let bPreventNav = false;


        if (this.oNextRoute) {

            let aoUnloadRoutes = null;
            let asLoadRoutes = null;
            let oPopState = null;

            console.log("updateRoute");
            console.log("Current loaded routes array")
            console.log(this.aoCurrentRoutesModules);

            console.log("Next Routes routes");
            console.log(this.oNextRoute.asRoutePaths);

            // Check to see if we already have modules loaded
            if (this.aoCurrentRoutesModules && this.aoCurrentRoutesModules.length) {

                // Loop throught the routes compare them to what is already loaded
                for (let r = 0, rLen = this.oNextRoute.asRoutePaths.length; r < rLen; r++) {

                    // Validate the current cursor has this index
                    if (this.aoCurrentRoutesModules[r]) {

                        // Check to see if we dont have a match, because that is when we care!
                        if (this.aoCurrentRoutesModules[r].sModulePath !== this.oNextRoute.asRoutePaths[r]) {
                            


                        }
                    }
                    else {

                        // Everything from this point on is new!
                        asLoadRoutes = this.oNextRoute.asRoutePaths.slice(r);

                    }

                    if ((r + 1 === rLen) && (this.aoCurrentRoutesModules.length > rLen)) {

                        console.log("At the end of the requested route, but there are more routes still in memory");

                        aoUnloadRoutes = this.aoCurrentRoutesModules.slice(rLen).reverse();

                        this.aoCurrentRoutesModules = this.aoCurrentRoutesModules.slice(0, rLen);
                    }

                }

            }
            else {

                // Nothing to worry about unloading, so load all new
                asLoadRoutes = this.oNextRoute.asRoutePaths;
            }

            /*
             * Screen to unloaded
             */

            // Check to see if we have a page that is being unloaded
            if (this.cCurrentScreenModule) {

                if (this.cCurrentScreenModule._unload) {

                    bPreventNav = this.cCurrentScreenModule._navGuard();

                    if(!bPreventNav){
                        oPopState = this.cCurrentScreenModule._unload();    
                    }
                }
                else {
                    
                    console.log("There is a screen, but no unload method to call");
                }

            }
            else {

                console.log("No screen module to unload");
            }

            if(!bPreventNav){

                /*
                 * Old route to unload (already in reversed order)
                 */

                // Check for modules that need to be unloaded
                if (aoUnloadRoutes && aoUnloadRoutes.length) {

                    for (let u = 0, uLen = aoUnloadRoutes.length; u < uLen; u++) {
                     
                        console.log(aoUnloadRoutes[u]);

                        if (aoUnloadRoutes[u].cModule._unload) {

                            await aoUnloadRoutes[u].cModule._unload();
                        }
                        
                    }
                }

                /*
                 * New route modules to load
                 */

                // Check for load modules
                if (asLoadRoutes && asLoadRoutes.length) {

                    for (let sRouteModule of asLoadRoutes) {
                        /*DEBUG: START*/
                        console.log('sRoute',sRouteModule);
                        /*DEBUG: END*/
                        let cModule = await this.loadModule(sRouteModule);

                        let oRouteProps = {};
                        if(this.aoCurrentRoutesModules){
                            oRouteProps.oParentRoute = this.aoCurrentRoutesModules[this.aoCurrentRoutesModules.length - 1].cModule;
                        }
                        oRouteProps.oNavState = oNavState;
                        oRouteProps.cAPP = this.cAPP;

                        let sRoutePath = sRouteModule.replace("/js/screens/", "");
                        // sRoutePath = sRoutePath.replace(/\//gi, "-")
                       
                        oRouteProps.sRoutePath = sRoutePath;

                        /*DEBUG: START*/
                        console.log('sRoutePath', sRoutePath);
                        /*DEBUG: END*/

                        // Build the module object
                        let oModuleLoad = {
                            sModulePath: sRouteModule,
                            cModule: new cModule.default(oRouteProps)
                        };

                        // Initialize the router modules
                        await oModuleLoad.cModule._init();

                        // Initialize the array if it doesnt exist
                        if (!this.aoCurrentRoutesModules) {
                            this.aoCurrentRoutesModules = [];
                        }

                        // Save off the module
                        this.aoCurrentRoutesModules.push(oModuleLoad);
                    }

                }

                /*DEBUG: START*/
                console.log('currentRoutes', this.aoCurrentRoutesModules);
                /*DEBUG: END*/

                /*
                 * Screen to load
                 */
                if (this.oNextRoute.sPagePath) {

                    let cScreen = await this.loadModule(this.oNextRoute.sPagePath);

                    /*DEBUG: START*/
                    console.log('cScreen', cScreen);
                    console.log('oNextRoute', this.oNextRoute)
                    /*DEBUG: END*/

                    let oScreenProps = {};
                    if(this.aoCurrentRoutesModules){
                        oScreenProps.oParentRoute = this.aoCurrentRoutesModules[this.aoCurrentRoutesModules.length - 1].cModule;
                    }
                   
                    oScreenProps.oNavState = oNavState;
                    oScreenProps.cAPP = this.cAPP;

                    this.cCurrentScreenModule = new cScreen.default(oScreenProps, this.cAPP);

                    await this.cCurrentScreenModule._init();

                }
                else {

                    throw new Error(`No new next route screen was defined`);
                }

                // Set history state of current URL with value returned to oPopState. 
                window.history.replaceState(oPopState, "",);

                // Create new history entry with new pages URL. 
                window.history.pushState({}, "", this.oNextRoute.sURLBar);
        
                document.title = this.cCurrentScreenModule.sTitle;
        
                // Clear out the next router info
                this.oNextRoute = null;
            }

        }
        else {

            throw new Error(`No new next route is currently defined`);
        }

    }

    validateRoute(sRequestedRoute, sRequestSource = "manual") {

        // Check to make sure the path is not
        if (!sRequestSource.startsWith('..')) {

            let bWasFlushed = false;

            let sWorkingPath = sRequestedRoute;

            let oNewCursor = this.oRoutes;
            let asRoutePaths = [];
            let sPagePath = null;
            let sURLBar = "/";
            let sFromRoot = "";

            if (sWorkingPath.startsWith('/')) {
                sWorkingPath = sWorkingPath.slice(1);
            }

            while (true) {

                let sBasePath = null;
                let routeModulePath = null;

                if (sFromRoot && sFromRoot.length) {
                    sBasePath = `${this.sScreenRoot}/${sFromRoot}`;
                }
                else {
                    sBasePath = this.sScreenRoot;
                }

                // A route folder request
                if (sWorkingPath.length === 0 || sWorkingPath === "index") {

                    // Check to see if index exists for this route
                    if (oNewCursor.index) {

                        // Check for and add a router module if one exists in this folder
                        if (oNewCursor['@routeModule']) {

                            routeModulePath = `${sBasePath}/${oNewCursor['@routeModule']}`;

                            if (asRoutePaths.indexOf(routeModulePath) === -1) {
                                asRoutePaths.push(routeModulePath);
                            }
                        }

                        // Build the page path
                        sPagePath = `${sBasePath}/${oNewCursor.index}`;

                        sURLBar += "index";

                        break;
                    }
                    else {

                        throw new Error(`Invalid route requested: ${sRequestedRoute}. Requested folder doesnt have a index.`);
                    }

                }
                else {

                    // Check to see if we have a path seperator
                    let iSlashIndex = sWorkingPath.indexOf('/');
                    let sPathPiece = null;

                    if (iSlashIndex === -1) {

                        sPathPiece = sWorkingPath.slice(0);
                        sWorkingPath = "";
                    }
                    else {

                        sPathPiece = sWorkingPath.slice(0, iSlashIndex);
                        sWorkingPath = sWorkingPath.slice(iSlashIndex + 1);
                    }

                    // Check the current path peice
                    if (oNewCursor[sPathPiece]) {

                        // Check to see if this is a page
                        if (typeof oNewCursor[sPathPiece] === "string") {

                            if (sWorkingPath.length === 0) {

                                // Check for and add a router module if one exists in this folder and is not part of the current path
                                if (oNewCursor['@routeModule']) {

                                    routeModulePath = `${sBasePath}/${oNewCursor['@routeModule']}`;

                                    if (asRoutePaths.indexOf(routeModulePath) === -1) {
                                        asRoutePaths.push(routeModulePath);
                                    }
                                }

                                // Build the page path
                                sPagePath = `${sBasePath}/${oNewCursor[sPathPiece]}`;

                                sURLBar += sPathPiece;

                                break;
                            }
                            else {

                                throw new Error(`Invalid route requested: ${sRequestedRoute}. Reached end of cursor, but not path`);
                            }

                        }
                        else {

                            // Check for and add a router module if one exists in this folder and is not part of the current path
                            if (oNewCursor['@routeModule']) {

                                routeModulePath = `${sBasePath}/${oNewCursor['@routeModule']}`;

                                if (asRoutePaths.indexOf(routeModulePath) === -1) {
                                    asRoutePaths.push(routeModulePath);
                                }
                            }

                            // Update from path to adopt the next level
                            if (sFromRoot.length) { 
                                sFromRoot = `${sFromRoot}/${sPathPiece}`;
                            }
                            else {
                                sFromRoot = sPathPiece;
                            }

                            sURLBar += `${sPathPiece}/`;

                            // Advance the cursor
                            oNewCursor = oNewCursor[sPathPiece];

                        }

                    }
                    else {

                        throw new Error(`Invalid route requested: ${sRequestedRoute}. Path cursor does not have a option for: ${sPathPiece}`);
                    }

                }

            }

            this.oNextRoute = {
                asRoutePaths: asRoutePaths,
                sPagePath: sPagePath,
                sURLBar: sURLBar
            }

        }
        else {

            throw new Error(`Route paths must be direct siblings or absolutes from route`);
        }

    }

    async navTo(sDestination, oNavData, eEvent, sSource, oHistoryState) {
        
        console.log(`User requested to navigate to: ${sDestination}`, eEvent, sSource, oNavData);

        let bContinue = true;

        // Start by trying to request the 
        try {

            this.validateRoute(sDestination);
        }
        catch(oError) {

            bContinue = false;
            console.log(oError);
        }

        if (bContinue) {

            let oNavState = {};
            if(sDestination){
                oNavState.sDestination = sDestination;
            }
            if(sSource){
                oNavState.sSource = sSource;
            }
            if(oNavData){
                oNavState.oNavData = oNavData;
            }        
            if(oHistoryState){
                oNavState.oHistoryState = oHistoryState;
            }
            

            this.updateRoute(oNavState);
        }

    }
}
